import abc
from magma.t import Direction, Type


class MagmaProtocolMeta(type):
    @abc.abstractmethod
    def _to_magma_(cls):
        # To retrieve underlying magma type.
        raise NotImplementedError()

    @abc.abstractmethod
    def _qualify_magma_(cls, direction: Direction):
        # To qualify underlying type (e.g. give me a Foo with the underlying
        # type qualified to be an input).
        raise NotImplementedError()

    @abc.abstractmethod
    def _flip_magma_(cls):
        # To flip underlying type (e.g. give me a Foo with the underlying type
        # flipped).
        raise NotImplementedError()

    def qualify(cls, direction: Direction):
        return cls._qualify_magma_(direction)

    @abc.abstractmethod
    def _from_magma_value_(cls, val: Type):
        # To create an instance from a value.
        raise NotImplementedError()

    def flip(cls):
        return cls._flip_magma_()


class MagmaProtocol(metaclass=MagmaProtocolMeta):
    @abc.abstractmethod
    def _get_magma_value_(self):
        # To access underlying magma value.
        raise NotImplementedError()

    @classmethod
    def is_clock(cls):
        return cls._to_magma_().is_clock()

    @classmethod
    def is_input(cls):
        return cls._to_magma_().is_input()

    @classmethod
    def is_mixed(cls):
        return cls._to_magma_().is_mixed()

    @classmethod
    def is_output(cls):
        return cls._to_magma_().is_output()

    def iswhole(self):
        return self._get_magma_value_().iswhole()

    def value(self):
        return self._get_magma_value_().value()

    def trace(self):
        return self._get_magma_value_().trace()

    def driven(self):
        return self._get_magma_value_().driven()

    def flatten(self):
        return self._get_magma_value_().flatten()

    def __iter__(self):
        return iter(self._get_magma_value_())

    @property
    def name(self):
        return self._get_magma_value_().name

    def wire(self, other):
        if isinstance(other, MagmaProtocol):
            other = other._get_magma_value_()
        self._get_magma_value_().wire(other)

    def __imatmul__(self, other):
        self.wire(other)



from ..array import Array
from magma.bit import Bit
from ..bits import Bits
from ..circuit import coreir_port_mapping
from ..conversions import as_bits, from_bits
from ..digital import Digital
from ..generator import Generator2
from ..interface import IO
from .passes import DefinitionPass
from ..ref import NamedRef, ArrayRef, PortViewRef
from ..t import In, Out
from ..tuple import Tuple


def _sanitize_name(name):
    return name.replace("[", "_").replace("]", "")


def _simulate_wire(self, value_store, state_store):
    value_store.set_value(self.O, value_store.get_value(self.I))


class Wire(Generator2):
    def __init__(self, T):
        if issubclass(T, Digital):
            coreir_lib = "corebit"
            coreir_genargs = None
            # Convert to bit so we wrap named types like clock if necessary
            T = Bit
        else:
            width = T.flat_length()
            T = Bits[width]
            coreir_lib = "coreir"
            coreir_genargs = {"width": width}
        self.name = "Wire"
        self.io = IO(I=In(T), O=Out(T))
        self.simulate = _simulate_wire
        self.coreir_name = "wire"
        self.coreir_lib = coreir_lib
        self.coreir_genargs = coreir_genargs
        self.renamed_ports = coreir_port_mapping


class InsertCoreIRWires(DefinitionPass):
    def __init__(self, main):
        super().__init__(main)
        self.seen = set()
        self.wire_map = {}

    def _make_wire(self, driver, value, definition):
        is_bits = (isinstance(driver.name, ArrayRef) and
                   isinstance(driver, Digital) and not driver.name.anon())

        # If this is a member of a bits, emit only one wire for the parent
        # Bits, since it shouldn't be flattened
        if is_bits:
            index = driver.name.index
            driver = driver.name.array

        T = type(driver)
        if driver not in self.wire_map:
            driver_name = driver.name.qualifiedname("_")
            driver_name = _sanitize_name(driver_name)

            # Rename driver so it doesn't conflict with new wire instance name
            if isinstance(driver.name, NamedRef):
                driver.name.name = "_" + driver.name.name

            name = f"{driver_name}"
            with definition.open():
                wire_inst = Wire(T)(name=name)
                self.wire_map[driver] = wire_inst
        wire_inst = self.wire_map[driver]
        wire_input = wire_inst.I
        wire_output = wire_inst.O

        if is_bits:
            # Select the index off the parent bits wire
            wire_input = wire_input[index]
            wire_output = wire_output[index]
            driver = driver[index]

        if not isinstance(driver, Digital):
            driver = as_bits(driver)
            wire_output = from_bits(T, wire_output)

        # Could be already wired for fanout cases
        if not wire_input.driven():
            if isinstance(wire_input, Array):
                for d, w in zip(driver, wire_input):
                    # Could have wired up child elements already
                    if not w.driven():
                        w @= d
            else:
                wire_input @= driver
        value @= wire_output

    def _insert_wire(self, value, definition):
        if value.is_mixed():  # mixed children
            for child in value:
                if not child.is_output():
                    self._insert_wire(child, definition)
            return
        if value in self.seen:
            return  # in the case of inouts, we may see more than once
        self.seen.add(value)
        driver = value.value()

        while (driver is not None and driver.name.anon() and
               not driver.is_output()):
            value, driver = driver, driver.value()

        descend = (isinstance(driver, (Array, Tuple)) and
                   not driver.iswhole())
        if descend:
            for child in value:
                self._insert_wire(child, definition)
            return

        if driver is None or driver.is_output() or driver.is_inout():
            return

        if isinstance(driver.name, PortViewRef):
            # Don't descend into hierarchical connections since we assume this
            # logic will be run in that context in another stage of the pass
            return

        value.unwire(driver)
        self._make_wire(driver, value, definition)
        self._insert_wire(driver, definition)

    def __call__(self, definition):
        if getattr(definition, "_coreir_wires_inserted_", False):
            return
        definition._coreir_wires_inserted_ = True
        # Copy instances because inserting wire will append to instances.
        instances_copy = definition.instances.copy()
        for instance in instances_copy:
            for value in instance.interface.ports.values():
                if not value.is_output():
                    self._insert_wire(value, definition)
        for value in definition.interface.ports.values():
            if not value.is_output():
                self._insert_wire(value, definition)



try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache

cachedFunctions = []


def clear_cachedFunctions():
    for func in cachedFunctions:
        func.cache_clear()


def cache_definition(fn):
    cachedFunctions.append(lru_cache(maxsize=None)(fn))
    return cachedFunctions[-1]

def singleton(cls):
    instance = [None]
    def wrapper(*args, **kwargs):
        if instance[0] is None:
            instance[0] = cls(*args, **kwargs)
        return instance[0]

    return wrapper

from magma.protocol_type import MagmaProtocolMeta, MagmaProtocol

# wires
from .wire import wire

# types
from .t import *

from .bit import *

from .array import *
from .bits import *
from .bfloat import BFloat
from .tuple import *
from .clock import *
from .clock_io import ClockIO
from .conversions import *
from .interface import *
from .ref import LazyDefnRef, LazyInstRef

# circuit
from .circuit import *

# higher-order operators
from .braid import *

# verilog
from .frontend.verilog import (declare_from_verilog, declare_from_verilog_file,
                               define_from_verilog, define_from_verilog_file,
                               DeclareFromVerilog, DeclareFromVerilogFile,
                               DefineFromVerilog, DefineFromVerilogFile,
                               set_verilog_importer)
from .backend.verilog import *

# compile
from .compile import *

from .logging import root_logger


mantle_target = None
def set_mantle_target(t):
     global mantle_target
     if mantle_target is not None and mantle_target != t:
         root_logger().warning('changing mantle target', mantle_target, t )
     mantle_target = t

from .backend.util import set_codegen_debug_info
from .enum import Enum
import magma.util
import magma.syntax

from .is_primitive import isprimitive
from .is_definition import isdefinition

from .uniquification import UniquificationPass


from hwtypes.bit_vector_abc import TypeFamily

from magma.family import get_family
_Family_ = get_family()

from .generator import Generator, Generator2
from .monitor import MonitorIO, MonitorGenerator, make_monitor_ports
from .inline_verilog import inline_verilog
from .display import display, posedge, negedge, File, time
from .log import info, debug, warning, error


from .syntax.sequential2 import sequential2
from .syntax.combinational2 import combinational2

from magma.primitives import LUT, Mux, mux, Register, slice, reduce



from typing import Union

import coreir
import hwtypes

from magma.bit import Bit
from magma.bits import Bits
from magma.circuit import coreir_port_mapping
from magma.conversions import as_bits, from_bits
from magma.interface import IO
from magma.generator import Generator2
from magma.t import Type, Kind, In, Out
from magma.clock import (AbstractReset, Reset, ResetN, AsyncReset, AsyncResetN,
                         Clock, Enable)
from magma.clock_io import ClockIO
from magma.primitives.mux import Mux


class _CoreIRRegister(Generator2):
    """
    Internally used generator for CoreIR register primitive
    """
    def __init__(self, width, init=0, has_async_reset=False,
                 has_async_resetn=False):
        self.name = "reg_P"
        self.coreir_configargs = {"init": coreir.type.BitVector[width](init)}
        T = Bits[width]
        self.io = IO(I=In(T), CLK=In(Clock), O=Out(T))

        if has_async_resetn and has_async_reset:
            raise ValueError("Cannot have posedge and negedge asynchronous "
                             "reset")

        if has_async_reset:
            self.io += IO(arst=In(AsyncReset))
            self.name += "R"
            self.coreir_configargs["arst_posedge"] = True

        if has_async_resetn:
            self.io += IO(arst=In(AsyncResetN))
            self.name += "R"
            self.coreir_configargs["arst_posedge"] = False

        self.stateful = True
        self.default_kwargs = {"init": coreir.type.BitVector[width](init)}
        self.coreir_genargs = {"width": width}
        self.renamed_ports = coreir_port_mapping

        self.coreir_name = "reg"
        if (has_async_reset or has_async_resetn):
            self.coreir_name += "_arst"

        self.verilog_name = "coreir_" + self.coreir_name
        self.coreir_lib = "coreir"

        def _simulate(self, value_store, state_store):
            cur_clock = value_store.get_value(self.CLK)

            if not state_store:
                state_store['prev_clock'] = cur_clock
                state_store['cur_val'] = hwtypes.BitVector[width](init)

            if has_async_reset or has_async_resetn:
                cur_reset = value_store.get_value(self.arst)

            prev_clock = state_store['prev_clock']
            clock_edge = cur_clock and not prev_clock

            new_val = state_store['cur_val'].as_bool_list()

            if clock_edge:
                new_val = value_store.get_value(self.I)

            if has_async_reset and cur_reset or \
                    has_async_resetn and not cur_reset:
                new_val = hwtypes.BitVector[width](init)

            state_store['prev_clock'] = cur_clock
            state_store['cur_val'] = hwtypes.BitVector[width](new_val)
            value_store.set_value(self.O, new_val)
        self.simulate = _simulate


class Register(Generator2):
    def __init__(self, T: Kind, init: Union[Type, int] = None, reset_type:
                 AbstractReset = None, has_enable: bool = False,
                 reset_priority=True):
        """
        T: The type of the value that is stored inside the register (e.g.
           Bits[5])

        init: (optional) A const value (i.e. init.const() == True) of type T or
              an int to be used as the initial value of the register.
              If no value is provided, the register will be initialized with 0

        reset_type: (optional) The type of the reset port (also specifies the
                    semantic behavior of the reset signal)

        reset_priority: (optional) boolean flag choosing whether synchronous
                        reset (RESET or RESETN) has priority over enable
        """
        if not isinstance(T, Kind):
            raise TypeError(
                f"Expected instance of Kind for argument T, not {type(T)}")
        if init is not None and not isinstance(init, (Type, int)):
            raise TypeError(
                f"Expected instance of Type or int for argument init, not "
                f"{type(init)}")
        if reset_type is not None and not issubclass(reset_type, AbstractReset):
            raise TypeError(
                f"Expected subclass of AbstractReset for argument reset_type, "
                f"not {type(reset_type)}")

        self.io = IO(I=In(T), O=Out(T))
        has_async_reset = reset_type == AsyncReset
        has_async_resetn = reset_type == AsyncResetN
        has_reset = reset_type == Reset
        has_resetn = reset_type == ResetN
        self.io += ClockIO(has_enable=has_enable,
                           has_async_reset=has_async_reset,
                           has_async_resetn=has_async_resetn,
                           has_reset=has_reset,
                           has_resetn=has_resetn)
        if init is None:
            init = 0

        if isinstance(init, int):
            init = T(init)

        if has_async_reset or has_async_resetn:
            coreir_init = int(as_bits(init))
        else:
            coreir_init = 0

        reg = _CoreIRRegister(T.flat_length(), init=coreir_init,
                              has_async_reset=has_async_reset,
                              has_async_resetn=has_async_resetn)()
        O = reg.O
        self.io.O @= from_bits(T, O)

        I = self.io.I
        if (has_reset or has_resetn) and has_enable:
            if has_reset:
                reset_port = self.io.RESET
            else:
                reset_port = self.io.RESETN

            if reset_priority:
                I = Mux(2, T)(name="enable_mux")(O, I, self.io.CE)
                I = Mux(2, T)()([I, init], reset_port)
            else:
                I = Mux(2, T)()([I, init], reset_port)
                I = Mux(2, T)(name="enable_mux")(O, I, self.io.CE)
        elif has_enable:
            I = Mux(2, T)(name="enable_mux")(O, I, self.io.CE)
        elif has_reset:
            I = Mux(2, T)()(I, init, self.io.RESET)
        reg.I @= as_bits(I)



import functools
import ast
import copy

import astor
from staticfg import CFGBuilder
from staticfg.builder import invert
from ast_tools.stack import inspect_enclosing_env

from .transforms import inline_yield_from_functions
from ..ast_utils import get_ast, compile_function_to_file
from ..bitutils import clog2


def is_if_not_true(node):
    return (
        isinstance(node, ast.UnaryOp)
        and isinstance(node.op, ast.Not)
        and isinstance(node.operand, ast.NameConstant)
        and node.operand.value is True
    )


class RemoveIfTrues(ast.NodeTransformer):
    """
    Remove `if True:` nodes by replacing them with their body
    Remove `if not True:` nodes by replacing them with `pass`
    """
    def visit_If(self, node):
        node = self.generic_visit(node)
        if isinstance(node.test, ast.NameConstant) and node.test.value is True:
            return node.body
        elif is_if_not_true(node.test):
            return ast.Pass()
        return node


class MergeInverseIf(ast.NodeTransformer):
    """
    Merge cases where
    ```
    if cond:
        ...
    if invert(cond):
        ...
    ```
    into
    ```
    if cond:
        ...
    else:
        ...
    ```

    `invert` comes from the `staticfg` library, which is used in marking
    control flow graph paths, so this just tries to reconstruct the if/else
    structure found in the original code based on the control flow graph
    construction
    """
    def visit(self, node):
        if hasattr(node, 'body'):
            node.body = self.visit_body(node.body)
        if hasattr(node, 'orelse'):
            node.orelse = self.visit_body(node.orelse)
        return self.generic_visit(node)

    def visit_body(self, body):
        new_body = []
        skip_next = False
        for i, node in enumerate(body):
            if skip_next:
                continue
            new_body.append(node)
            if i == len(body) - 1:
                break
            if (isinstance(node, ast.If) and isinstance(body[i + 1], ast.If)
                    and ast.dump(invert(node.test))
                    == ast.dump(body[i + 1].test)):
                node.orelse = body[i + 1].body
                skip_next = True
            else:
                skip_next = False
        return new_body


def is_yield(statement):
    """
    Is this a statement of the form `yield <expr>`
    """
    return (isinstance(statement, ast.Expr)
            and isinstance(statement.value, ast.Yield))


def collect_paths_to_yield(start_idx, block):
    """
    Given a start idx in a block, collect the paths to the next yield.

    Use a start offset because some yields occur inside a basic block, so the
    start of a path may not be the start of the block.

    Copies branch nodes and stores their `exitcase` into the test for use later
    (so we know the required condition of this path)
    """
    path = []
    for statement in block.statements[start_idx:]:
        path.append(statement)
        if is_yield(statement):
            return [path]
    paths = []
    for exit in block.exits:
        _path = path
        if is_if_not_true(exit.exitcase):
            # skip 'if not True:'
            continue
        elif exit.exitcase is not None:
            assert (len(_path) == 1
                    and isinstance(_path[0], (ast.If, ast.While))), \
                "Expect branch"
            _path = _path.copy()
            _path[0] = copy.copy(_path[0])
            _path[0].test = exit.exitcase
        paths.extend(_path + p for p in collect_paths_to_yield(0, exit.target))
    return paths


def build_method_name_map(tree):
    name_map = {}
    for stmt in tree.body:
        if isinstance(stmt, ast.FunctionDef):
            name_map[stmt.name] = stmt
    return name_map


def get_options(tree):
    class_vars = {}
    for stmt in tree.body:
        if isinstance(stmt, ast.Assign):
            assert len(stmt.targets) == 1 and \
                isinstance(stmt.targets[0], ast.Name)
            class_vars[stmt.targets[0].id] = stmt.value
    return class_vars


def _coroutine(defn_env, fn):
    tree = get_ast(fn).body[0]
    method_name_map = build_method_name_map(tree)

    options = get_options(tree)
    manual_encoding = options.get("_manual_encoding_", ast.NameConstant(False))
    if not isinstance(manual_encoding, ast.NameConstant):
        raise TypeError("_manual_encoding_ should be a boolean literal")
    manual_encoding = manual_encoding.value

    reset_type = options.get("_reset_type_", ast.parse("m.AsyncReset").body[0])

    call_method = method_name_map["__call__"]
    call_method = inline_yield_from_functions(call_method, method_name_map)

    # build cfg
    cfg = CFGBuilder().build(call_method.name, call_method)
    # debug cfg visualizer
    # cfg.build_visual(tree.name, 'pdf')

    # map from yield ast node (by id) to paths to other yields
    yield_paths = {}
    # map from yield ast node (by id) to numeric id (for state encoding)
    yield_encoding_map = {}

    # populate maps
    for block in cfg:
        for i, statement in enumerate(block.statements):
            if is_yield(statement):
                paths = collect_paths_to_yield(i + 1, block)
                yield_paths[statement] = paths
                if not manual_encoding:
                    # Encode by index
                    encoding = len(yield_encoding_map)
                else:
                    # TODO: Assumes previous stmt sets encoding
                    encoding = astor.to_source(
                        block.statements[i - 1].value
                    ).rstrip()
                yield_encoding_map[statement] = encoding

    # add yield state register declaration
    yield_state_width = clog2(len(yield_encoding_map))
    if not manual_encoding:
        method_name_map["__init__"].body.append(ast.parse(
            f"self.yield_state: m.Bits[{yield_state_width}] = 0"
        ).body[0])

    # create new call body
    call_method.body = []
    for i, (yield_, paths) in enumerate(yield_paths.items()):
        # condition based on yield state encoding for current start yield
        test = ast.parse(
            f"self.yield_state == {yield_encoding_map[yield_]}"
        ).body[0].value

        if i == 0:
            # for first if, populate new call method body
            curr_body = []
            call_method.body.append(ast.If(test, curr_body, []))
            prev_if = call_method.body[0]
        elif i == len(yield_paths) - 1:
            # Use the previous if's orelse block for final state
            curr_body = prev_if.orelse
        else:
            # use the previous if's orelse block (so we generate an if/elif
            # chain)
            curr_body = []
            prev_if.orelse.append(ast.If(test, curr_body, []))
            prev_if = prev_if.orelse[-1]

        # Emite code for each path
        for path in paths:
            body = curr_body
            end_yield = None
            for statement in path:
                if is_yield(statement):
                    end_yield = statement
                    break
                body.append(copy.deepcopy(statement))
                # If we encounter a branch, subsequent statements go inside the
                # branch
                if isinstance(statement, ast.While):
                    body[-1] = ast.If(body[-1].test, [], [])
                if isinstance(statement, ast.If):
                    body[-1].body = []
                    body[-1].orelse = []
                if isinstance(statement, (ast.If, ast.While)):
                    body = body[-1].body
            assert end_yield is not None, "Found path not ending in yield"

            if not manual_encoding:
                # Finish with updating the yield state register with the end
                # yield of the path
                body.append(ast.parse(
                    f"self.yield_state = m.bits("
                    f"{yield_encoding_map[end_yield]}, {yield_state_width})"
                ).body[0])

            # Return the value of the original end yield
            body.append(ast.Return(end_yield.value.value))

    # Remove if trues for readability
    call_method = RemoveIfTrues().visit(call_method)
    # Merge `if cond:` with `if not cond` for readability
    call_method = MergeInverseIf().visit(call_method)

    # Sequential stage
    reset_type_str = astor.to_source(reset_type).rstrip()
    tree.decorator_list = [ast.parse(
        f"m.circuit.sequential(reset_type={reset_type_str})"
    ).body[0].value]

    # print(astor.to_source(tree))
    circuit = compile_function_to_file(tree, tree.name, defn_env)
    return circuit


coroutine = inspect_enclosing_env(_coroutine)



from hwtypes.bit_vector_abc import TypeFamily
import magma as m


_Family_ = None


def get_family():
    global _Family_
    if _Family_ is None:
        _Family_ = TypeFamily(m.Bit, m.Bits, m.UInt, m.SInt)
    return _Family_



from collections import OrderedDict
import hwtypes
import pyverilog.dataflow.visit
import pyverilog.vparser.parser
from ..array import Array
from ..bit import Bit
from ..bits import Bits
from ..circuit import Circuit
from ..interface import IO
from ..math import log2_ceil
from ..passes.tsort import tsort
from ..t import In, Out, InOut
from .verilog_importer import (ImportMode, MultipleModuleDeclarationError,
                               MultiplePortDeclarationError, VerilogImporter,
                               VerilogImportError)
from .verilog_utils import int_const_str_to_int


class PyverilogImportError(VerilogImportError):
    pass


def _evaluate_node(node, params):
    """Evaluates the pyverilog node @node"""
    if isinstance(node, pyverilog.vparser.ast.IntConst):
        return int_const_str_to_int(node.value)
    if isinstance(node, pyverilog.vparser.ast.Rvalue):
        return _evaluate_node(node.var, params)
    if isinstance(node,
                  (pyverilog.vparser.ast.Minus, pyverilog.vparser.ast.Uminus)):
        l = _evaluate_node(node.left, params)
        r = _evaluate_node(node.right, params)
        return l - r
    if isinstance(node, pyverilog.vparser.ast.Plus):
        l = _evaluate_node(node.left, params)
        r = _evaluate_node(node.right, params)
        return l + r
    if isinstance(node, pyverilog.vparser.ast.Divide):
        l = _evaluate_node(node.left, params)
        r = _evaluate_node(node.right, params)
        return l // r  # floor division
    if isinstance(node, pyverilog.vparser.ast.Times):
        l = _evaluate_node(node.left, params)
        r = _evaluate_node(node.right, params)
        return l * r  # floor division
    if isinstance(node, pyverilog.vparser.ast.Identifier):
        return params[node.name]
    if isinstance(node, pyverilog.vparser.ast.SystemCall):
        syscall = node.syscall
        if syscall == "clog2":
            operand = _evaluate_node(node.children()[0], params)
            return log2_ceil(operand)
        raise PyverilogImportError(
            f"Unsupported verilog system call: {syscall}")
    raise PyverilogImportError(f"Unsupported expression: {type(node)}")


def _get_width(width, param_map):
    """Evaluates width.msb, width.lsb and returns their difference"""
    msb = _evaluate_node(width.msb, param_map)
    lsb = _evaluate_node(width.lsb, param_map)
    return msb - lsb + 1


def _parse_port(port):
    """
    Returns the pair (name: str, io: Union[(Input|Output|Inout), None]), where
    the second element is the pyverilog encoded type of the input, or None (if
    the port list name-only.
    """
    if isinstance(port, pyverilog.vparser.parser.Ioport):
        io = port.first
        return (io.name, io)
    if isinstance(port, pyverilog.vparser.parser.Port):
        return (port.name, None)
    # Not expected to reach here.
    raise PyverilogImportError(f"Unsupported port type: {type(port)}")


def _magma_type(io, params):
    """
    Returns the magma type corresponding to @io, which is a pyverilog Input,
    Output, or Inout.
    """
    direction = InOut
    if isinstance(io, pyverilog.vparser.parser.Input):
        direction = In
    elif isinstance(io, pyverilog.vparser.parser.Output):
        direction = Out

    typ = Bit
    if io.width is not None:
        width = _get_width(io.width, params)
        typ = Bits[width]

    # Generate multidimensional arrays if necessary. Note that we guard with
    # hasattr to make this backwards compatible.
    if hasattr(io, "dimensions") and io.dimensions is not None:
        for length in reversed(io.dimensions.lengths):
            width = _get_width(length, params)
            typ = Array[width, typ]

    return direction(typ)


def _get_lines(src, start, end):
    return "\n".join(src.split("\n")[start - 1:end])


class _ModuleGraphVisitor(pyverilog.dataflow.visit.NodeVisitor):
    """
    This implementation of pyverilog's NodeVisitor collects all module
    definitions in the root node.
    """
    def __init__(self):
        self.defns = OrderedDict()

    def visit_ModuleDef(self, defn):
        if defn.name in self.defns:
            raise MultipleModuleDeclarationError(defn.name)
        self.defns[defn.name] = defn


class PyverilogImporter(VerilogImporter):
    """Implementation of VerilogImporter using pyverilog"""
    def _import_defn(self, defn, mode):
        ports = {}
        default_params = {}
        # Parse module declaration for parameters (and default values) and port
        # declarations.
        for param_list in defn.paramlist.params:
            for param in param_list.list:
                default_params[param.name] = _evaluate_node(
                    param.value, default_params)
        for port in defn.portlist.ports:
            name, io = _parse_port(port)
            if name in ports:
                raise MultiplePortDeclarationError(name)
            ports[name] = io
        # Check that the port declaration is not mixed-style, i.e. either all
        # declared in the module declaration, or all declared in the body.
        is_io = (p is not None for p in ports.values())
        if not all(is_io) and any(is_io):
            raise MixedPortDeclarationsError()
        # Iterate on the body of the moudle definition for paramters (and
        # default values) and port declarations.
        for child in defn.children():
            if not isinstance(child, pyverilog.vparser.parser.Decl):
                continue
            for subchild in child.children():
                if isinstance(subchild, pyverilog.vparser.parser.Parameter):
                    default_params[subchild.name] = _evaluate_node(
                        subchild.value, default_params)
                elif isinstance(subchild, (pyverilog.vparser.parser.Input,
                                           pyverilog.vparser.parser.Output,
                                           pyverilog.vparser.parser.Inout)):
                    ports[subchild.name] = subchild
        # Convert pyverilog ports to magma types and define a magma circuit.
        decl = {n: self.map_type(n, _magma_type(port, default_params))
                for n, port in ports.items()}

        class _FromVerilog(Circuit):
            name = defn.name
            io = IO(**decl)

        return _FromVerilog, default_params

    def import_(self, src, mode):
        parser = pyverilog.vparser.parser.VerilogParser()
        ast = parser.parse(src)
        visitor = _ModuleGraphVisitor()
        visitor.visit(ast)
        for name, defn in visitor.defns.items():
            circ, default_params = self._import_defn(defn, mode)
            if mode is ImportMode.DEFINE:
                circ.verilogFile = _get_lines(src, defn.lineno, defn.end_lineno)
                circ.verilog_source = src
            circ.coreir_config_param_types = {
                k: type(v)
                for k, v in default_params.items()
            }
            circ.default_kwargs = default_params
            self.add_module(circ)



from magma.bits import Bits
from magma.conversions import array


def slice(value: Bits, start: Bits, width: int):
    """
    Dynamic slice based off the value of `start` with constant `width`.

    Example:
    ```
    class TestSlice(m.Circuit):
        # Slice a 6 bit window of I using x as the start index
        IO = m.IO(
            I=m.In(m.Bits[10]),
            x=m.In(m.Bits[2]),
            O=m.Out(m.Bits[6])
        )

        io.O @= m.slice(io.I, start=io.x, width=6)
    ```

    Notes:
      * For now we only support slicing bits with bits index
      * We could add support for slicing Arrays
      * We could restrict index to be a UInt
    """
    # Construct an array where the index `i` is the slice of bits from `i` to
    # `i+width`, index into this array using `start`.
    return array([value[i:i + width] for i in range(len(value) - width)])[start]



from .passes import EditDefinitionPass
from ..is_definition import isdefinition


def _drive_if_undriven_input(port):
    if port.is_mixed():
        # list comp so it doesn't short circuit
        undrivens = [_drive_if_undriven_input(p) for p in port]
        return any(undrivens)
    if port.is_input() and port.trace() is None:
        port.undriven()
        return True
    return False


def _drive_undriven(interface):
    undriven = False
    for port in interface.ports.values():
        undriven |= _drive_if_undriven_input(port)
    return undriven


class DriveUndrivenPass(EditDefinitionPass):
    def edit(self, circuit):
        if _drive_undriven(circuit.interface):
            circuit._is_definition = True
        if not isdefinition(circuit):
            return
        for inst in circuit.instances:
            _drive_undriven(inst.interface)



from typing import Optional

from ast_tools.stack import SymbolTable
from ast_tools.passes import apply_ast_passes, ssa, if_to_phi, bool_to_bit

from ..circuit import Circuit, IO
from ..clock_io import ClockIO
from ..t import Type

from .util import build_io_args, build_call_args, wire_call_result


class _SequentialListWrapper(list):
    def __getitem__(self, i):
        result = super().__getitem__(i)
        if isinstance(result, list):
            return _SequentialListWrapper(result)
        return result

    def __setitem__(self, i, value):
        item = self[i]
        if not isinstance(item, Circuit):
            raise TypeError("Excepted setitem target to be a Circuit")

        if not isinstance(value, (Circuit, int, Type)):
            raise TypeError("Excepted setitem value to be a Circuit, Type, or "
                            f"int, not {type(value)}")

        item(value)


def sequential_getattribute(self, key):
    """
    Wrap lists so we can use the setattr interface with lists of registers
    """
    result = object.__getattribute__(self, key)
    if isinstance(result, list):
        return _SequentialListWrapper(result)
    return result


def sequential_setattr(self, key, value):
    # TODO: for now we assume this is a register, ideally we'd type check this,
    # but we need to have a notion of a register primitive (e.g. right now the
    # mantle reg wraps the coreir reg primitive, so technically the register is
    # an arbitrary user-defined circuit)

    # We can at least match the value is a circuit with outputs that match the
    # input of the attribute circuit
    if not isinstance(getattr(self, key), Circuit):
        raise TypeError("Excepted setattr key to be a Circuit")

    if not isinstance(value, (Circuit, int, Type)):
        raise TypeError("Excepted setattr value to be a Circuit, Type, or int",
                        f"not {type(value)}")

    # Simply use function call syntax for now (should automatically retrieve
    # the output of value)
    getattr(self, key)(value)


def _process_phi_arg(arg):
    if isinstance(arg, Circuit):
        # Call with no args to retrieve output(s)
        arg = arg()
        if isinstance(arg, list):
            # TODO: We could support tuples/products
            raise TypeError("Cannot use circuit with multiple outputs as "
                            "argument to phi")
    return arg


def _seq_phi(s, t, f):
    # Unpack registers/circuits used in getattr syntax
    t, f = map(_process_phi_arg, (t, f))
    return s.ite(t, f)


def sequential2(pre_passes=[], post_passes=[],
                debug: bool = False,
                env: Optional[SymbolTable] = None,
                path: Optional[str] = None,
                file_name: Optional[str] = None,
                **clock_io_kwargs):
    """ clock_io_kwargs used for ClockIO params, e.g. async_reset """
    passes = (pre_passes +
              [ssa(strict=False), bool_to_bit(), if_to_phi(_seq_phi)] +
              post_passes)

    def seq_inner(cls):
        cls.__call__ = apply_ast_passes(passes, debug=debug, env=env,
                                        path=path,
                                        file_name=file_name)(cls.__call__)

        if "self" in cls.__call__.__annotations__:
            raise Exception("Assumed self did not have annotation")

        io_args = build_io_args(cls.__call__.__annotations__)

        class SequentialCircuit(Circuit):
            name = cls.__name__
            io = IO(**io_args) + ClockIO(**clock_io_kwargs)
            call_args = [cls()]

            # Monkey patch setattribute for register assign syntax, we could
            # also add this in a Sequential base class, but if we do that we
            # might as well use a metaclass rather than a decorator, but to
            # preserve the old interface we do this for now
            cls.__setattr__ = sequential_setattr

            # also need to patch getattr to support list attributes (see
            # test_sequential2:test_sequential2_pre_unroll)
            cls.__getattribute__ = sequential_getattribute

            call_args += build_call_args(io, cls.__call__.__annotations__)

            call_result = cls.__call__(*call_args)
            wire_call_result(io, call_result, cls.__call__.__annotations__)
        return SequentialCircuit
    return seq_inner



from .t import Direction, In
from .digital import DigitalMeta, Digital
from .wire import wire
from magma.bit import Bit
from magma.array import Array
from magma.tuple import Tuple


class _ClockType(Digital):
    @classmethod
    def is_clock(cls):
        return True

    def unused(self):
        Bit.unused(self)

    def undriven(self):
        Bit.undriven(self)


class Clock(_ClockType, metaclass=DigitalMeta):
    pass


ClockIn = Clock[Direction.In]
ClockOut = Clock[Direction.Out]


class AbstractReset(_ClockType, metaclass=DigitalMeta):
    pass


# synchronous reset, active high (i.e. reset when signal is 1)
class Reset(AbstractReset):
    pass


ResetIn = Reset[Direction.In]
ResetOut = Reset[Direction.Out]


# synchronous reset, active low (i.e. reset when signal is 0)
class ResetN(AbstractReset):
    pass


ResetNIn = ResetN[Direction.In]
ResetNOut = ResetN[Direction.Out]


# asynchronous reset, active high (i.e. reset when signal is 1)
class AsyncReset(AbstractReset):
    pass


AsyncResetIn = AsyncReset[Direction.In]
AsyncResetOut = AsyncReset[Direction.Out]


# asynchronous reset, active low (i.e. reset when signal is 0)
class AsyncResetN(AbstractReset):
    pass


AsyncResetNIn = AsyncResetN[Direction.In]
AsyncResetNOut = AsyncResetN[Direction.Out]


# Preset
# Clear
class Enable(_ClockType, metaclass=DigitalMeta):
    pass


EnableIn = Enable[Direction.In]
EnableOut = Enable[Direction.Out]

ClockTypes = (Clock, Reset, AsyncReset, AsyncResetN, Enable)


def ClockInterface(has_enable=False, has_reset=False, has_ce=False,
                   has_async_reset=False, has_async_resetn=False):
    args = ['CLK', In(Clock)]
    has_enable |= has_ce
    if has_enable:
        args += ['CE', In(Enable)]
    if has_reset:
        args += ['RESET', In(Reset)]
    if has_async_reset:
        args += ['ASYNCRESET', In(AsyncReset)]
    if has_async_resetn:
        args += ['ASYNCRESETN', In(AsyncResetN)]
    return args


def wire_clock_port(port, clocktype, defnclk):
    wired = False
    if isinstance(port, Tuple):
        for elem in port:
            wired |= wire_clock_port(elem, clocktype, defnclk)
    elif isinstance(port, Array):
        wired = wire_clock_port(port[0], clocktype, defnclk)
        # Only traverse all children circuit if first child has a clock
        if not wired:
            return False
        # TODO: Magma doesn't support length zero array, so slicing a
        # length 1 array off the end doesn't work as expected in normal
        # Python, so we explicilty slice port.ts
        for t in port.ts[1:]:
            for elem in port[1:]:
              wire_clock_port(elem, clocktype, defnclk)
    elif isinstance(port, clocktype) and not port.driven():
        wire(defnclk, port)
        wired = True
    return wired


def first(iterable):
    return next((item for item in iterable if item is not None), None)


def get_first_clock(port, clocktype):
    if isinstance(port, clocktype):
        return port
    if isinstance(port, Tuple):
        clks = (get_first_clock(elem, clocktype) for elem in port)
        return first(clks)
    if isinstance(port, Array):
        return get_first_clock(port[0], clocktype)
    return None


def wireclocktype(defn, inst, clocktype):
    # Check common case: top level clock port
    clks = (port if isinstance(port, clocktype) else None
            for port in defn.interface.ports.values())
    defnclk = first(clks)
    if defnclk is None:
        # Check recursive types
        clks = (get_first_clock(port, clocktype)
                for port in defn.interface.ports.values())
        defnclk = first(clks)
    if defnclk is None:
        return
    for port in inst.interface.inputs(include_clocks=True):
        wire_clock_port(port, clocktype, defnclk)


def wiredefaultclock(defn, inst):
    wireclocktype(defn, inst, Clock)


def wireclock(define, circuit):
    wireclocktype(define, circuit, Reset)
    wireclocktype(define, circuit, AsyncReset)
    wireclocktype(define, circuit, AsyncResetN)
    wireclocktype(define, circuit, Enable)



import operator
import functools

from hwtypes import BitVector

from magma.bit import Bit
from magma.bits import Bits
from magma.circuit import coreir_port_mapping, IO
from magma.generator import Generator2
from magma.t import In, Out


def _reduce_factory(coreir_name, operator):
    class Reduce(Generator2):
        def __init__(self, width: int):
            self.io = IO(I=In(Bits[width]), O=Out(Bit))
            self.coreir_lib = "coreir"
            self.coreir_name = coreir_name
            self.name = f"coreir_{coreir_name}_{width}"
            self.coreir_genargs = {"width": width}
            self.renamed_ports = coreir_port_mapping
            self.primitive = True
            self.stateful = False

            def simulate(self, value_store, state_store):
                I = BitVector[width](value_store.get_value(self.I))
                O = functools.reduce(operator, I.bits())
                value_store.set_value(self.O, O)

            self.simulate = simulate
    return Reduce


_OP_MAP = {
    operator.and_: _reduce_factory("andr", operator.and_),
    operator.or_: _reduce_factory("orr", operator.or_),
    operator.xor: _reduce_factory("xorr", operator.xor)
}


def reduce(operator, bits: Bits):
    if not isinstance(bits, Bits):
        raise TypeError("m.reduce only works with Bits")
    if operator not in _OP_MAP:
        raise ValueError(f"{operator} not supported")
    return _OP_MAP[operator](len(bits))()(bits)
